@@ -1081,7 +1081,6 @@ pem_CreateObject
 )
 {
     CK_OBJECT_CLASS objClass;
-    CK_BBOOL isToken;
     NSSCKFWSlot *fwSlot;
     CK_SLOT_ID slotID;
     CK_BBOOL cacert;
@@ -1090,27 +1089,11 @@ pem_CreateObject
     int nobjs = 0;
     int i;
     long objid;
-#if 0
-    pemToken *token;
-#endif
     int cipher = 0;
     char *ivstring = NULL;
     pemInternalObject *listObj = NULL;
     pemObjectListItem *listItem = NULL;
 
-    /*
-     * only create token objects
-     */
-    isToken = pem_GetBoolAttribute(CKA_TOKEN, pTemplate,
-                                   ulAttributeCount, pError);
-    if (CKR_OK != *pError) {
-        return (NSSCKMDObject *) NULL;
-    }
-    if (!isToken) {
-        *pError = CKR_ATTRIBUTE_VALUE_INVALID;
-        return (NSSCKMDObject *) NULL;
-    }
-
     /* What slot are we adding the object to? */
     fwSlot = NSSCKFWSession_GetFWSlot(fwSession);
     if ((NSSCKFWSlot *) NULL == fwSlot) {
@@ -1120,10 +1103,6 @@ pem_CreateObject
     }
     slotID = NSSCKFWSlot_GetSlotID(fwSlot);
 
-#if 0
-    token = (pemToken *) mdToken->etc;
-#endif
-
     /*
      * only create keys and certs.
      */
@@ -1163,8 +1142,8 @@ pem_CreateObject
     }
 
     if (objClass == CKO_CERTIFICATE) {
-        nobjs = ReadDERFromFile(&derlist, filename, /* ascii */ PR_TRUE,
-                                &cipher, &ivstring, /* certs only */ PR_TRUE);
+        nobjs = ReadDERFromFile(&derlist, filename, &cipher, &ivstring,
+                                /* certs only */ PR_TRUE);
         if (nobjs < 1)
             goto loser;
 
@@ -1208,8 +1187,8 @@ pem_CreateObject
         SECItem certDER;
         PRBool added;
 
-        nobjs = ReadDERFromFile(&derlist, filename, /* ascii */ PR_TRUE,
-                                &cipher, &ivstring, /* keys only */ PR_FALSE);
+        nobjs = ReadDERFromFile(&derlist, filename, &cipher, &ivstring,
+                                /* keys only */ PR_FALSE);
         if (nobjs < 1)
             goto loser;
 
