@@ -45,6 +45,7 @@
 #include <nssb64.h>
 #include <nssbase.h>
 #include <secerr.h>
+#include <secitem.h>
 #include <secpkcs7.h>
 
 #include <stdarg.h>
@@ -65,59 +66,37 @@ static int put_object(SECItem *der, SECItem ***derlist, int *count)
     return 0;
 }
 
-static SECItem *AllocItem(SECItem * item, unsigned int len)
-{
-    SECItem *result = NULL;
-
-    if (item == NULL) {
-	result = NSS_ZAlloc(NULL, sizeof(SECItem));
-	if (result == NULL) {
-	    goto loser;
-	}
-    } else {
-	PORT_Assert(item->data == NULL);
-	result = item;
-    }
-
-    result->len = len;
-    if (len) {
-        result->data = NSS_ZAlloc(NULL, len);
-    }
-
-    return (result);
-
-  loser:
-    return (NULL);
-}
-
 static SECStatus FileToItem(SECItem * dst, PRFileDesc * src)
 {
-    PRFileInfo info;
-    PRInt32 numBytes;
-    PRStatus prStatus;
+    static const PRInt32 chunk = 65536;
+    PRInt32 bytesReadTotal = 0;
 
-    if (!dst)
-	return SECFailure;
+    for (;;) {
+	PRInt32 bytesReadNow;
+	PRInt32 newSize = bytesReadTotal + chunk;
+	if (newSize < chunk)
+	    /* int overflow */
+	    break;
 
-    prStatus = PR_GetOpenFileInfo(src, &info);
+	if (SECITEM_ReallocItemV2(NULL, dst, newSize) != SECSuccess)
+	    /* out of memory */
+	    break;
 
-    if (prStatus != PR_SUCCESS || info.type == PR_FILE_DIRECTORY) {
-	return SECFailure;
+	bytesReadNow = PR_Read(src, dst->data + bytesReadTotal, chunk);
+	if (bytesReadNow < 0)
+	    /* read error */
+	    break;
+
+	if (bytesReadNow == 0) {
+	    /* EOF */
+	    dst->len = bytesReadTotal;
+	    return SECSuccess;
+	}
+
+	bytesReadTotal += bytesReadNow;
     }
 
-    /* XXX workaround for 3.1, not all utils zero dst before sending */
-    dst->data = 0;
-    if (!AllocItem(dst, info.size+1))
-	goto loser;
-
-    numBytes = PR_Read(src, dst->data, info.size);
-    if (numBytes != info.size) {
-	goto loser;
-    }
-
-    return SECSuccess;
-  loser:
-    NSS_ZFreeIf(dst->data);
+    free(dst->data);
     return SECFailure;
 }
 
@@ -145,8 +124,8 @@ static SECStatus ConvertAsciiToZAllocItem(SECItem *der, const char *ascii)
 }
 
 /* returns count of objects read, or -1 on error */
-int ReadDERFromFile(SECItem *** derlist, char *filename, PRBool ascii,
-		    int *cipher, char **ivstring, PRBool certsonly)
+int ReadDERFromFile(SECItem *** derlist, char *filename, int *cipher,
+		    char **ivstring, PRBool certsonly)
 {
     SECStatus rv;
     PRFileDesc *inFile;
@@ -156,144 +135,133 @@ int ReadDERFromFile(SECItem *** derlist, char *filename, PRBool ascii,
     SECItem filedata;
     char *c, *iv;
 
-    filedata.data = NULL;
+    memset(&filedata, 0, sizeof filedata);
 
     inFile = PR_Open(filename, PR_RDONLY, 0);
     if (!inFile)
 	return -1;
 
-    if (ascii) {
-	/* First convert ascii to binary */
-	char *asc, *body;
+    /* First convert ascii to binary */
+    char *asc, *body;
 
-	/* Read in ascii data */
-	rv = FileToItem(&filedata, inFile);
-	if (rv != SECSuccess) {
-	    PR_Close(inFile);
-	    return -1;
-	}
-	asc = (char *) filedata.data;
-	if (!asc) {
-	    PR_Close(inFile);
-	    return -1;
-	}
+    /* Read in ascii data */
+    rv = FileToItem(&filedata, inFile);
+    if (rv != SECSuccess) {
+	PR_Close(inFile);
+	return -1;
+    }
+    asc = (char *) filedata.data;
+    if (!asc) {
+	PR_Close(inFile);
+	return -1;
+    }
 
-	/* check for headers and trailers and remove them */
-	if (strstr(asc, "-----BEGIN") != NULL) {
-            int key = 0;
-	    while ((asc) && ((body = strstr(asc, "-----BEGIN")) != NULL)) {
-                key = 0;
-		if ((strncmp(body, "-----BEGIN RSA PRIVATE KEY", 25) == 0) ||
-		    (strncmp(body, "-----BEGIN PRIVATE KEY", 21) == 0)) {
-                    key = 1;
-		    c = body;
+    /* check for headers and trailers and remove them */
+    if (strstr(asc, "-----BEGIN") != NULL) {
+	int key = 0;
+	while ((asc) && ((body = strstr(asc, "-----BEGIN")) != NULL)) {
+	    key = 0;
+	    if ((strncmp(body, "-----BEGIN RSA PRIVATE KEY", 25) == 0) ||
+		(strncmp(body, "-----BEGIN PRIVATE KEY", 21) == 0)) {
+		key = 1;
+		c = body;
+		body = strchr(body, '\n');
+		if (NULL == body)
+		    goto loser;
+		body++;
+		if (strncmp(body, "Proc-Type: 4,ENCRYPTED", 22) == 0) {
 		    body = strchr(body, '\n');
 		    if (NULL == body)
 			goto loser;
 		    body++;
-		    if (strncmp(body, "Proc-Type: 4,ENCRYPTED", 22) == 0) {
-			body = strchr(body, '\n');
-			if (NULL == body)
+		    if (strncmp(body, "DEK-Info: ", 10) == 0) {
+			body += 10;
+			c = body;
+			body = strchr(body, ',');
+			if (body == NULL)
+			    goto loser;
+			*body = '\0';
+			if (!strcasecmp(c, "DES-EDE3-CBC"))
+			    *cipher = NSS_DES_EDE3_CBC;
+			else if (!strcasecmp(c, "DES-CBC"))
+			    *cipher = NSS_DES_CBC;
+			else {
+			    *cipher = -1;
 			    goto loser;
-			body++;
-			if (strncmp(body, "DEK-Info: ", 10) == 0) {
-			    body += 10;
-			    c = body;
-			    body = strchr(body, ',');
-			    if (body == NULL)
-				goto loser;
-			    *body = '\0';
-			    if (!strcasecmp(c, "DES-EDE3-CBC"))
-				*cipher = NSS_DES_EDE3_CBC;
-			    else if (!strcasecmp(c, "DES-CBC"))
-				*cipher = NSS_DES_CBC;
-			    else {
-				*cipher = -1;
-				goto loser;
-			    }
-			    body++;
-			    iv = body;
-			    body = strchr(body, '\n');
-			    if (body == NULL)
-				goto loser;
-			    *body = '\0';
-			    body++;
-			    *ivstring = PORT_Strdup(iv);
 			}
-		    } else {	/* Else the private key is not encrypted */
-			*cipher = 0;
-			body = c;
+			body++;
+			iv = body;
+			body = strchr(body, '\n');
+			if (body == NULL)
+			    goto loser;
+			*body = '\0';
+			body++;
+			*ivstring = PORT_Strdup(iv);
 		    }
+		} else {	/* Else the private key is not encrypted */
+		    *cipher = 0;
+		    body = c;
 		}
-		der = NSS_ZAlloc(NULL, sizeof(SECItem));
-                if (der == NULL)
-                    goto loser;
-
-		char *trailer = NULL;
-		asc = body;
-		body = strchr(body, '\n');
-		if (!body)
-		    body = strchr(asc, '\r');	/* maybe this is a MAC file */
-		if (body) {
-		    trailer = strstr(++body, "-----END");
-		}
-		if (trailer != NULL) {
-		    asc = trailer + 1;
-		    *trailer = '\0';
-		} else {
-		    goto loser;
-		}
-
-		/* Convert to binary */
-		rv = ConvertAsciiToZAllocItem(der, body);
-		if (rv) {
-		    goto loser;
-		}
-                if ((certsonly && !key) || (!certsonly && key)) {
-		    error = put_object(der, derlist, &count);
-		    if (error) {
-			goto loser;
-		    }
-		    der = NULL;
-                } else {
-                    NSS_ZFreeIf(der->data);
-                    NSS_ZFreeIf(der);
-		    der = NULL;
-                }
-	    }			/* while */
-	} else {		/* No headers and footers, translate the blob */
+	    }
 	    der = NSS_ZAlloc(NULL, sizeof(SECItem));
 	    if (der == NULL)
 		goto loser;
 
-	    rv = ConvertAsciiToZAllocItem(der, asc);
+	    char *trailer = NULL;
+	    asc = body;
+	    body = strchr(body, '\n');
+	    if (!body)
+		body = strchr(asc, '\r');	/* maybe this is a MAC file */
+	    if (body) {
+		trailer = strstr(++body, "-----END");
+	    }
+	    if (trailer != NULL) {
+		asc = trailer + 1;
+		*trailer = '\0';
+	    } else {
+		goto loser;
+	    }
+
+	    /* Convert to binary */
+	    rv = ConvertAsciiToZAllocItem(der, body);
 	    if (rv) {
 		goto loser;
 	    }
-	    /* NOTE: This code path has never been tested. */
-	    error = put_object(der, derlist, &count);
-	    if (error) {
-		goto loser;
+	    if ((certsonly && !key) || (!certsonly && key)) {
+		error = put_object(der, derlist, &count);
+		if (error) {
+		    goto loser;
+		}
+		der = NULL;
+	    } else {
+		NSS_ZFreeIf(der->data);
+		NSS_ZFreeIf(der);
+		der = NULL;
 	    }
-	    der = NULL;
-	}
+	}			/* while */
+    } else {		/* No headers and footers, translate the blob */
+	der = NSS_ZAlloc(NULL, sizeof(SECItem));
+	if (der == NULL)
+	    goto loser;
 
-	NSS_ZFreeIf(filedata.data);
-        filedata.data = 0;
-	filedata.len = 0;
-    } else {
-	/* Read in binary der */
-	rv = FileToItem(der, inFile);
-	if (rv != SECSuccess) {
-	    PR_Close(inFile);
-	    return -1;
+	rv = ConvertAsciiToZAllocItem(der, asc);
+	if (rv) {
+	    goto loser;
 	}
+	/* NOTE: This code path has never been tested. */
+	error = put_object(der, derlist, &count);
+	if (error) {
+	    goto loser;
+	}
+	der = NULL;
     }
+
+    free(filedata.data);
     PR_Close(inFile);
     return count;
 
   loser:
-    NSS_ZFreeIf(filedata.data);
+    free(filedata.data);
     if (der) {
 	NSS_ZFreeIf(der->data);
 	NSS_ZFreeIf(der);
